#include <libspe2.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>


/*
 * Embedded SPE program 
 */
extern spe_program_handle_t spe_program;


/*
 * Size of the matrices 
 */
#define N 64


/*
 * The matrices in main memory (aligned to quadword boundary)
 */
static float A[N*N] __attribute__((aligned(16)));
static float B[N*N] __attribute__((aligned(16)));
static float C[N*N] __attribute__((aligned(16)));
static float Ccopy[N*N] __attribute__((aligned(16))); // for error checking


/*
 * Structure used for communicating control information with SPE.
 * Must be multiple of 16 bytes large (hence the padding).
 */
typedef struct
{
   float *pA;
   float *pB;
   float *pC;
   unsigned int ticks;
   unsigned int padding;
} program_data;


/*
 * Generates [-1, 1] uniform random float
 */
float random_float()
{
   return 2.f * (((float) rand()) / RAND_MAX - 0.5f);
}


/*
 * Naive SGEMM used for error checking.
 */
void sgemm()
{
   int i, j, k;
   for( j = 0; j < N; j++ ) {
      for( i = 0; i < N; i++ ) {
         for( k = 0; k < N; k++ ) {
            Ccopy[i + j*N] += A[k + i*N] * B[k + j*N];
         }
      }
   }
}


/*
 * Copies a N*N contiguous block from src to dst
 */
void copy_block(float *src, float *dst)
{
   int i;
   for( i = 0; i < N*N; i++ )
      dst[i] = src[i];
}


/*
 * Pthread function for PPE SPE thread
 */
void *spe_thread_function(void *arg)
{
   unsigned int EP = SPE_DEFAULT_ENTRY;
   spe_context_ptr_t *speid = (spe_context_ptr_t *) arg;
   program_data pd __attribute__((aligned(16)));
   int i, j;
   float sec;
   float max_err = 0.f;
   float err;
   float P, Ppeak = 25.6f;

   /*
    * Generate random matrices.
    */
   printf(">> Generating random matrices A, B, C........."); fflush(stdout);
   for( j = 0; j < N; j++ ) {
      for( i = 0; i < N; i++ ) {
         A[i + j*N] = random_float();
         B[i + j*N] = random_float();
         C[i + j*N] = random_float();
      }
   }
   printf("OK\n");

   /*
    * Make a copy of C for error checking
    */
   printf(">> Making copy of C..........................."); fflush(stdout); 
   copy_block(C, Ccopy);
   printf("OK\n");

   /*
    * Compute the correct solution using naive impl. on PPE
    */
   printf(">> Computing expected solution on PPE........."); fflush(stdout);
   sgemm();
   printf("OK\n");

   /*
    * Run the SPE context 
    */
   printf(">> Running SGEMM on SPE......................."); fflush(stdout);
   pd.pA = &A[0];
   pd.pB = &B[0];
   pd.pC = &C[0];
   spe_context_run(*speid, &EP, 0, &pd, NULL, NULL);
   printf("OK\n");

   /*
    * Compute the maximum componentwise error
    */
   printf(">> Computing error in solution................"); fflush(stdout);
   for( i = 0; i < N; i++ ) {
      for( j = 0; j < N; j++ ) {
         err = C[i + j*N] - Ccopy[i + j*N];
         err = fabsf(err);
         if( err > max_err) max_err = err;
      }
   }
   printf("OK\n");

   /*
    * Print result to stdout
    */
   printf(">> DONE!\n\n");
   sec = ((float) pd.ticks) / 26666666;
   P = 2.E-9f*N*N*N/sec;
   printf("Results:\n----------------------------------------\n");
   printf("Ticks:    %10d\n", pd.ticks);
   printf("Seconds:  %10.6f\n", sec);
   printf("GFLOP/s:  %10.6f  (%.2f%% of peak)\n", P, P/Ppeak*100.f);
   printf("max(|e|): %10.3e  %s\n", max_err, (max_err > 1E-4) ? "(warning)" : "(ok)");

   /*
    * Exit gracefully
    */
   pthread_exit(NULL);
}


/*
 * PPE Main entry point
 */
int main(void)
{
   pthread_t my_thread;
   spe_context_ptr_t speid;
   int retval;

   /*
    * Init randomizer
    */
   srand(time(NULL));

   /*
    * Create SPE context
    */
   speid = spe_context_create(SPE_MAP_PS, NULL);

   /*
    * Load SPE executable into SPE context
    */
   spe_program_load(speid, &spe_program);

   /*
    * Create pthread to run SPE context (send context pointer as argument)
    */
   retval = pthread_create(&my_thread, NULL, spe_thread_function, (void*) &speid);
   if( retval ) {
      printf("pthread create failed\n");
      exit(1);
   }
   

   /*
    * Synchronize with the SPE pthread 
    */
   retval = pthread_join(my_thread, NULL);
   if( retval ) {
      printf("pthread join failed\n");
      exit(1);
   }

   /* 
    * Destroy the context
    */
   spe_context_destroy(speid);

   return 0;
}
